/**********************************************************************
  sample04.c:
  
   This program is used for benchmarking the performance of OpenFFT with timing and GLOPS results.
   It can be executed with an arbitrary number of processes.
   Time is measured by OpenFFT.    
   A numeric input parameter can be provided for specifying the size of the 3 dimensions. If no input parameter is provided, it will be executed with a default size of 64^3 data points. 

***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "openfft.h"

int main(int argc, char* argv[])
{ 
  dcomplex *Rhor, *Rhok;
  double time0,TStime,TEtime,time1,time2,time3,time,time11;
  double timer0,timer1,timer2,timer3,timeF,timeC,timeOthers;
  int numprocs,myid;
  int N1,N2,N3,measure_time,print_memory,My_Max_NumGrid,My_NumGrid_In,My_NumGrid_Out;
  int i,j,k,l;
  double factor,err,err_all,n1,flops;
  int NTEST=10;
  int My_Index_In[6],My_Index_Out[6];

  /* MPI */
  MPI_Init(&argc, &argv); 
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD,&myid);

  /* Set global input grid size based on the program's parameters */

  if(argc==2){
    N1 = N2 = N3 = atoi(argv[1]);
  }
  else if(argc==4){
    N1 = atoi(argv[1]);
    N1 = atoi(argv[2]);
    N3 = atoi(argv[3]);
  }
  else{
    N1 = N2 = N3 = 64;
  }

  /* Set whether to use the timing and print memory functions of OpenFFT or not. Default=0 (not use) */

  measure_time = 1;
  print_memory = 0;

  /* Initialize OpenFFT */ 

  time1 = openfft_initialize(N1,N2,N3,measure_time,print_memory,
			     &My_Max_NumGrid,&My_NumGrid_In,My_Index_In,&My_NumGrid_Out,My_Index_Out);

  /* Allocate local input and output arrays */

  Rhor = (dcomplex*)malloc(sizeof(dcomplex)*My_Max_NumGrid); 
  Rhok = (dcomplex*)malloc(sizeof(dcomplex)*My_Max_NumGrid); 

  /* Set local input */
  
  for (i=0; i<My_NumGrid_In; i++){
    Rhor[i].r = 0.1; 
    Rhor[i].i = 0.2;
  }

  /* Repeat FFT transform for NTEST times */

  MPI_Barrier(MPI_COMM_WORLD);
  time2 = 0.0;
  timer0 = timer1 = timer2 = 0.0;

  for(l=0;l<NTEST;l++){
    time2  += openfft_execute(Rhor, Rhok);
    timer0 += OFFT_timers[0];
    timer1 += OFFT_timers[1];
    timer2 += OFFT_timers[2];

    /* Re-set local input */

    for (i=0; i<My_NumGrid_In; i++){
      Rhor[i].r = 0.1; 
      Rhor[i].i = 0.2;
    }

  }
  
  /* Average time of all processes */

  MPI_Allreduce(&time2,&time,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  time = time/numprocs;
  MPI_Allreduce(&timer0,&timeF,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  timeF = timeF/numprocs;
  MPI_Allreduce(&timer1,&timeC,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  timeC = timeC/numprocs;
  MPI_Allreduce(&timer2,&timeOthers,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  timeOthers = timeOthers/numprocs;
  MPI_Allreduce(&time1,&time11,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  time1 = time11/numprocs;

  /* for error checking */

  err = 0.0;
  for (i=0; i<My_NumGrid_Out; i++){
    err = err + Rhok[i].r + Rhok[i].i;
  }
  MPI_Allreduce(&err,&err_all,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  err_all = err_all/(N1*N2*N3);

  /* Calculate flops */

  n1 = (double)(N1*N2*N3);
  n1 = pow(n1,1.0/3.0);
  flops = 5.0*n1*log(n1)/log(2);
  flops = flops*3*pow(n1,2);
  flops = flops/(time/NTEST);

  /* Finalize OpenFFT */

  time3 = openfft_finalize();
  MPI_Allreduce(&time3,&time11,1,MPI_DOUBLE,MPI_SUM,MPI_COMM_WORLD);
  time3 = time11/numprocs;

  /* Free arrays */

  free(Rhor);
  free(Rhok);

  /* Print results */

  if(myid==0){
    printf("=======OpenFFT complex2complex========\n");
    printf("N1=%5d,N2=%5d,N3=%5d,numprocs=%5d\n",N1,N2,N3,numprocs);
    printf("Time openfft_execute Total  = %15.12f\n",time);
    printf("Time openfft_execute FFT    = %15.12f\n",timeF);
    printf("Time openfft_execute Comm   = %15.12f\n",timeC);
    printf("Time openfft_execute Others = %15.12f\n",timeOthers);
    printf("Gflops                      = %15.12f\n",flops/pow(1000,3));
    printf("Time openfft_initialize     = %15.12f\n",time1);
    printf("Time openfft_finalize       = %15.12f\n",time3);
    printf("Err_all                     = %15.12f\n",err_all);
  }

  MPI_Finalize();

}

