/*******************************************************
  openfft_initialize is used to prepare data for the transformation 
     Input: 3 dimensions of data: N1, N2, N3, measure_time and print_memory (0: disabled, 1: enabled)
     Output: arrays allocated and variables initialized
             My_Max_NumGrid: the maximum number of grid points allocated to a process
             My_NumGrid_In : the number of grid points allocated to a process for the first transformation
             My_Index_In   : the indexes of grid points allocated to a process for the first transformation
             My_NumGrid_Out: the number of grid points allocated to a process for the last transformation
             My_Index_Out  : the indexes of grid points allocated to a process for the last transformation
******************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "openfft.h"

double openfft_initialize(int N1, int N2, int N3, int measure_time, int print_memory,
			  int *My_Max_NumGrid, 
			  int *My_NumGrid_In, int *My_Index_In,
			  int *My_NumGrid_Out, int *My_Index_Out)
{

  int numprocs,myid,tag=999,ID,IDS,IDR;
  unsigned long long int AN,BN,CN,DN;
  unsigned long long int B_AB2,Bs,Be;
  unsigned long long int BN_AB,BN_CB,BN_CA;
  unsigned long long int GN,GNs,GR,n2D,N2D;
  unsigned long long int GN_AB,GN_CB,GN_CA;
  int n1,n2,n3;
  MPI_Status stat;
  MPI_Request request;
  double time0,TStime,TEtime;
  int myStart,myEnd;
  double MB;
  int size_array0,size_array1;
  MPI_Comm mpi_comm_level1 = MPI_COMM_WORLD;
  
  time0 = 0.0;

  MPI_Comm_size(mpi_comm_level1,&numprocs);
  MPI_Comm_rank(mpi_comm_level1,&myid);

  OFFT_measure_time = measure_time;
  OFFT_print_memory = print_memory;

  if(OFFT_measure_time){
      MPI_Barrier(mpi_comm_level1);
      openfft_dtime(&TStime);
  }

  OFFT_alloc_first = 1;
  OFFT_alloc_first1= 1;

  OFFT_Ngrid1 = N1;
  OFFT_Ngrid2 = N2;
  OFFT_Ngrid3 = N3;

  Max_OneD_Grids = OFFT_Ngrid1 > OFFT_Ngrid2 ? OFFT_Ngrid1 : OFFT_Ngrid2;
  Max_OneD_Grids =  Max_OneD_Grids > OFFT_Ngrid3 ?  Max_OneD_Grids : OFFT_Ngrid3;

  /****************************************************
    allocation of arrays:
  ****************************************************/

  OFFT_in  = fftw_malloc(sizeof(fftw_complex)*Max_OneD_Grids); 
  OFFT_out = fftw_malloc(sizeof(fftw_complex)*Max_OneD_Grids); 

  OFFT_p_C = fftw_plan_dft_1d(OFFT_Ngrid3,OFFT_in,OFFT_out,-1,FFTW_ESTIMATE);
  OFFT_p_B = fftw_plan_dft_1d(OFFT_Ngrid2,OFFT_in,OFFT_out,-1,FFTW_ESTIMATE);
  OFFT_p_A = fftw_plan_dft_1d(OFFT_Ngrid1,OFFT_in,OFFT_out,-1,FFTW_ESTIMATE);


  /****************************************************************
      AB to CA for MPI communication  
  ****************************************************************/

  /* set My_NumGrid_AB */

  N2D = OFFT_Ngrid1*OFFT_Ngrid2;
  My_NumGrid_AB =  (((myid+1)*N2D+numprocs-1)/numprocs)*OFFT_Ngrid3
                  - ((myid*N2D+numprocs-1)/numprocs)*OFFT_Ngrid3;
  OFFT_My_NumGrid_In = My_NumGrid_AB; 

  /* set myStart,myEnd,OFFT_My_Index_In */

  N2D = OFFT_Ngrid1*OFFT_Ngrid2;
  myStart = ((myid*N2D+numprocs-1)/numprocs)*OFFT_Ngrid3;
  myEnd   = (((myid+1)*N2D+numprocs-1)/numprocs)*OFFT_Ngrid3 - 1;

  N2D = OFFT_Ngrid2*OFFT_Ngrid3;
  My_Index_In[0] = myStart/N2D;
  My_Index_In[1] = (myStart-My_Index_In[0]*N2D)/OFFT_Ngrid3;
  My_Index_In[2] = myStart-My_Index_In[0]*N2D-My_Index_In[1]*OFFT_Ngrid3;
  
  My_Index_In[3] = myEnd/N2D;
  My_Index_In[4] = (myEnd-My_Index_In[3]*N2D)/OFFT_Ngrid3;
  My_Index_In[5] = myEnd-My_Index_In[3]*N2D-My_Index_In[4]*OFFT_Ngrid3;

  /* set GNs */

  N2D = OFFT_Ngrid1*OFFT_Ngrid2;
  GNs = ((myid*N2D+numprocs-1)/numprocs)*OFFT_Ngrid3;

  /* find Num_Snd_Grid_AB2CA */

  Num_Snd_Grid_AB2CA = (int*)malloc(sizeof(int)*numprocs);
  Num_Rcv_Grid_AB2CA = (int*)malloc(sizeof(int)*numprocs);

  for (ID=0; ID<numprocs; ID++) Num_Snd_Grid_AB2CA[ID] = 0;

  N2D = OFFT_Ngrid3*OFFT_Ngrid1; /* for CA */

  for (BN_AB=0; BN_AB<My_NumGrid_AB; BN_AB++){

    GN_AB = BN_AB + GNs;

    n1 = GN_AB/(OFFT_Ngrid2*OFFT_Ngrid3);
    n2 = (GN_AB - n1*(OFFT_Ngrid2*OFFT_Ngrid3))/OFFT_Ngrid3;
    n3 = GN_AB - n1*(OFFT_Ngrid2*OFFT_Ngrid3) - n2*OFFT_Ngrid3;
    n2D = n3*OFFT_Ngrid1 + n1;
    ID = n2D*numprocs/N2D;
    Num_Snd_Grid_AB2CA[ID]++;
  }

  /* MPI: Num_Snd_Grid_AB2CA */  

  for (ID=0; ID<numprocs; ID++){

    IDS = (myid + ID) % numprocs;
    IDR = (myid - ID + numprocs) % numprocs;

    tag = 999;
    MPI_Isend(&Num_Snd_Grid_AB2CA[IDS], 1, MPI_INT, IDS, tag, mpi_comm_level1, &request);
    MPI_Recv(&Num_Rcv_Grid_AB2CA[IDR], 1, MPI_INT, IDR, tag, mpi_comm_level1, &stat);
    MPI_Wait(&request,&stat);
  }

  /* allocation of arrays */  

  if (OFFT_alloc_first==0){

    for (ID=0; ID<numprocs; ID++){
      free(Index_Snd_Grid_AB2CA[ID]);
    }  
    free(Index_Snd_Grid_AB2CA);
  
    for (ID=0; ID<numprocs; ID++){
      free(Index_Rcv_Grid_AB2CA[ID]);
    }  
    free(Index_Rcv_Grid_AB2CA);

    free(ID_NN_AB2CA_S);
    free(ID_NN_AB2CA_R);
    free(GP_AB2CA_S);
    free(GP_AB2CA_R);
  }

  size_Index_Snd_Grid_AB2CA = 0;
  Index_Snd_Grid_AB2CA = (int**)malloc(sizeof(int*)*numprocs);
  for (ID=0; ID<numprocs; ID++){
    Index_Snd_Grid_AB2CA[ID] = (int*)malloc(sizeof(int)*Num_Snd_Grid_AB2CA[ID]);
    size_Index_Snd_Grid_AB2CA += Num_Snd_Grid_AB2CA[ID];
  }  
  
  size_Index_Rcv_Grid_AB2CA = 0; 
  Index_Rcv_Grid_AB2CA = (int**)malloc(sizeof(int*)*numprocs);
  for (ID=0; ID<numprocs; ID++){
    Index_Rcv_Grid_AB2CA[ID] = (int*)malloc(sizeof(int)*Num_Rcv_Grid_AB2CA[ID]);
    size_Index_Rcv_Grid_AB2CA += Num_Rcv_Grid_AB2CA[ID]; 
  }  

  OFFT_alloc_first = 0;

  /* construct Index_Snd_Grid_AB2CA */  

  for (ID=0; ID<numprocs; ID++) Num_Snd_Grid_AB2CA[ID] = 0;

  N2D = OFFT_Ngrid3*OFFT_Ngrid1;  /* for CA */

  for (BN_AB=0; BN_AB<My_NumGrid_AB; BN_AB++){

    GN_AB = BN_AB + GNs;

    n1 = GN_AB/(OFFT_Ngrid2*OFFT_Ngrid3);
    n2 = (GN_AB - n1*(OFFT_Ngrid2*OFFT_Ngrid3))/OFFT_Ngrid3;
    n3 = GN_AB - n1*(OFFT_Ngrid2*OFFT_Ngrid3) - n2*OFFT_Ngrid3;
    n2D = n3*OFFT_Ngrid1 + n1;
    ID = n2D*numprocs/N2D;
    GN_CA = n3*OFFT_Ngrid1*OFFT_Ngrid2 + n1*OFFT_Ngrid2 + n2;
    BN_CA = GN_CA - ((ID*N2D+numprocs-1)/numprocs)*OFFT_Ngrid2;
    Index_Snd_Grid_AB2CA[ID][Num_Snd_Grid_AB2CA[ID]] = BN_CA;
    Num_Snd_Grid_AB2CA[ID]++;
  }

  /* MPI: Index_Snd_Grid_AB2CA */

  for (ID=0; ID<numprocs; ID++){

    IDS = (myid + ID) % numprocs;
    IDR = (myid - ID + numprocs) % numprocs;

    tag = 999;
   
    if (Num_Snd_Grid_AB2CA[IDS]!=0){
      MPI_Isend( &Index_Snd_Grid_AB2CA[IDS][0], Num_Snd_Grid_AB2CA[IDS], 
		 MPI_INT, IDS, tag, mpi_comm_level1, &request);
    }

    if (Num_Rcv_Grid_AB2CA[IDR]!=0){
      MPI_Recv( &Index_Rcv_Grid_AB2CA[IDR][0], Num_Rcv_Grid_AB2CA[IDR], 
		MPI_INT, IDR, tag, mpi_comm_level1, &stat);
    }

    if (Num_Snd_Grid_AB2CA[IDS]!=0)  MPI_Wait(&request,&stat);
  }

  /* reset: Index_Snd_Grid_AB2CA

  Index_Snd_Grid_AB2CA:  BN_AB
  Index_Rcv_Grid_AB2CA:  BN_CA
  */

  for (ID=0; ID<numprocs; ID++) Num_Snd_Grid_AB2CA[ID] = 0;

  N2D = OFFT_Ngrid3*OFFT_Ngrid1; /* for CA */

  for (BN_AB=0; BN_AB<My_NumGrid_AB; BN_AB++){

    GN_AB = BN_AB + GNs;
    n1 = GN_AB/(OFFT_Ngrid2*OFFT_Ngrid3);
    n2 = (GN_AB - n1*(OFFT_Ngrid2*OFFT_Ngrid3))/OFFT_Ngrid3;
    n3 = GN_AB - n1*(OFFT_Ngrid2*OFFT_Ngrid3) - n2*OFFT_Ngrid3;
    n2D = n3*OFFT_Ngrid1 + n1;
    ID = n2D*numprocs/N2D;
    Index_Snd_Grid_AB2CA[ID][Num_Snd_Grid_AB2CA[ID]] = BN_AB;
    Num_Snd_Grid_AB2CA[ID]++;
  }

  /* find the maximum Num_Snd_Grid_AB2CA and Num_Rcv_Grid_AB2CA */

  Max_Num_Snd_Grid_AB2CA = 0;
  Max_Num_Rcv_Grid_AB2CA = 0;

  for (ID=0; ID<numprocs; ID++){
    if (Max_Num_Snd_Grid_AB2CA<Num_Snd_Grid_AB2CA[ID]){ 
      Max_Num_Snd_Grid_AB2CA = Num_Snd_Grid_AB2CA[ID];
    }

    if (Max_Num_Rcv_Grid_AB2CA<Num_Rcv_Grid_AB2CA[ID]){ 
      Max_Num_Rcv_Grid_AB2CA = Num_Rcv_Grid_AB2CA[ID];
    }
  }

  /* find NN_AB2CA_S and NN_AB2CA_R 
     and set ID_NN_AB2CA_S, 
             ID_NN_AB2CA_R,
             GP_AB2CA_S,
             GP_AB2CA_R
  */

  NN_AB2CA_S = 0;
  NN_AB2CA_R = 0;
 
  for (ID=0; ID<numprocs; ID++){
    IDS = (myid + ID) % numprocs;
    IDR = (myid - ID + numprocs) % numprocs;
    if (Num_Snd_Grid_AB2CA[IDS]!=0) NN_AB2CA_S++;
    if (Num_Rcv_Grid_AB2CA[IDR]!=0) NN_AB2CA_R++;
  }

  ID_NN_AB2CA_S = (int*)malloc(sizeof(int)*NN_AB2CA_S);
  ID_NN_AB2CA_R = (int*)malloc(sizeof(int)*NN_AB2CA_R);
  GP_AB2CA_S = (int*)malloc(sizeof(int)*(NN_AB2CA_S+1));
  GP_AB2CA_R = (int*)malloc(sizeof(int)*(NN_AB2CA_R+1));

  NN_AB2CA_S = 0;
  NN_AB2CA_R = 0;
  GP_AB2CA_S[0] = 0;
  GP_AB2CA_R[0] = 0;

  for (ID=0; ID<numprocs; ID++){

    IDS = (myid + ID) % numprocs;
    IDR = (myid - ID + numprocs) % numprocs;

    if (Num_Snd_Grid_AB2CA[IDS]!=0){
      ID_NN_AB2CA_S[NN_AB2CA_S] = IDS;
      NN_AB2CA_S++;
      GP_AB2CA_S[NN_AB2CA_S] = GP_AB2CA_S[NN_AB2CA_S-1] + Num_Snd_Grid_AB2CA[IDS];
    }

    if (Num_Rcv_Grid_AB2CA[IDR]!=0){
      ID_NN_AB2CA_R[NN_AB2CA_R] = IDR;
      NN_AB2CA_R++;
      GP_AB2CA_R[NN_AB2CA_R] = GP_AB2CA_R[NN_AB2CA_R-1] + Num_Rcv_Grid_AB2CA[IDR];
    }
  }

  /* set My_NumGrid_CA */

  N2D = OFFT_Ngrid3*OFFT_Ngrid1;
  My_NumGrid_CA = (((myid+1)*N2D+numprocs-1)/numprocs)*OFFT_Ngrid2
                  - ((myid*N2D+numprocs-1)/numprocs)*OFFT_Ngrid2;

  /****************************************************************
      CA to CB for MPI communication  
  ****************************************************************/

  /* set GNs */

  N2D = OFFT_Ngrid3*OFFT_Ngrid1;
  GNs = ((myid*N2D+numprocs-1)/numprocs)*OFFT_Ngrid2;

  /* find Num_Snd_Grid_CA2CB */

  Num_Snd_Grid_CA2CB = (int*)malloc(sizeof(int)*numprocs);
  Num_Rcv_Grid_CA2CB = (int*)malloc(sizeof(int)*numprocs);

  for (ID=0; ID<numprocs; ID++) Num_Snd_Grid_CA2CB[ID] = 0;

  N2D = OFFT_Ngrid3*OFFT_Ngrid2; /* for CB */

  for (BN_CA=0; BN_CA<My_NumGrid_CA; BN_CA++){

    GN_CA = GNs + BN_CA;    
    n3 = GN_CA/(OFFT_Ngrid1*OFFT_Ngrid2);
    n1 = (GN_CA - n3*(OFFT_Ngrid1*OFFT_Ngrid2))/OFFT_Ngrid2;
    n2 = GN_CA - n3*(OFFT_Ngrid1*OFFT_Ngrid2) - n1*OFFT_Ngrid2;
    n2D = n3*OFFT_Ngrid2 + n2;
    ID = n2D*numprocs/N2D;
    Num_Snd_Grid_CA2CB[ID]++;
  }

  /* MPI: Num_Snd_Grid_CA2CB */  

  for (ID=0; ID<numprocs; ID++){

    IDS = (myid + ID) % numprocs;
    IDR = (myid - ID + numprocs) % numprocs;

    tag = 999;
    MPI_Isend(&Num_Snd_Grid_CA2CB[IDS], 1, MPI_INT, IDS, tag, mpi_comm_level1, &request);
    MPI_Recv(&Num_Rcv_Grid_CA2CB[IDR], 1, MPI_INT, IDR, tag, mpi_comm_level1, &stat);
    MPI_Wait(&request,&stat);
  }

  /* allocation of arrays */  

  if (OFFT_alloc_first1==0){

    for (ID=0; ID<numprocs; ID++){
      free(Index_Snd_Grid_CA2CB[ID]);
    }  
    free(Index_Snd_Grid_CA2CB);
  
    for (ID=0; ID<numprocs; ID++){
      free(Index_Rcv_Grid_CA2CB[ID]);
    }  
    free(Index_Rcv_Grid_CA2CB);

    free(ID_NN_CA2CB_S);
    free(ID_NN_CA2CB_R);
    free(GP_CA2CB_S);
    free(GP_CA2CB_R);
  }

  size_Index_Snd_Grid_CA2CB = 0;
  Index_Snd_Grid_CA2CB = (int**)malloc(sizeof(int*)*numprocs);
  for (ID=0; ID<numprocs; ID++){
    Index_Snd_Grid_CA2CB[ID] = (int*)malloc(sizeof(int)*Num_Snd_Grid_CA2CB[ID]);
    size_Index_Snd_Grid_CA2CB += Num_Snd_Grid_CA2CB[ID];
  }  
  
  size_Index_Rcv_Grid_CA2CB = 0; 
  Index_Rcv_Grid_CA2CB = (int**)malloc(sizeof(int*)*numprocs);
  for (ID=0; ID<numprocs; ID++){
    Index_Rcv_Grid_CA2CB[ID] = (int*)malloc(sizeof(int)*Num_Rcv_Grid_CA2CB[ID]);
    size_Index_Rcv_Grid_CA2CB += Num_Rcv_Grid_CA2CB[ID]; 
  }  

  OFFT_alloc_first1 = 0;

  /* construct Index_Snd_Grid_CA2CB */  

  for (ID=0; ID<numprocs; ID++) Num_Snd_Grid_CA2CB[ID] = 0;

  N2D = OFFT_Ngrid3*OFFT_Ngrid2; /* for CB */

  for (BN_CA=0; BN_CA<My_NumGrid_CA; BN_CA++){

    GN_CA = GNs + BN_CA;    
    n3 = GN_CA/(OFFT_Ngrid1*OFFT_Ngrid2);
    n1 = (GN_CA - n3*(OFFT_Ngrid1*OFFT_Ngrid2))/OFFT_Ngrid2;
    n2 = GN_CA - n3*(OFFT_Ngrid1*OFFT_Ngrid2) - n1*OFFT_Ngrid2;
    n2D = n3*OFFT_Ngrid2 + n2;
    ID = n2D*numprocs/N2D;
    GN_CB = n3*OFFT_Ngrid2*OFFT_Ngrid1 + n2*OFFT_Ngrid1 + n1;
    BN_CB = GN_CB - ((ID*N2D+numprocs-1)/numprocs)*OFFT_Ngrid1;
    Index_Snd_Grid_CA2CB[ID][Num_Snd_Grid_CA2CB[ID]] = BN_CB;
    Num_Snd_Grid_CA2CB[ID]++;
  }

  /* MPI: Index_Snd_Grid_CA2CB */

  for (ID=0; ID<numprocs; ID++){

    IDS = (myid + ID) % numprocs;
    IDR = (myid - ID + numprocs) % numprocs;

    tag = 999;
   
    if (Num_Snd_Grid_CA2CB[IDS]!=0){
      MPI_Isend( &Index_Snd_Grid_CA2CB[IDS][0], Num_Snd_Grid_CA2CB[IDS], 
                 MPI_INT, IDS, tag, mpi_comm_level1, &request);
    }

    if (Num_Rcv_Grid_CA2CB[IDR]!=0){
      MPI_Recv( &Index_Rcv_Grid_CA2CB[IDR][0], Num_Rcv_Grid_CA2CB[IDR], 
                MPI_INT, IDR, tag, mpi_comm_level1, &stat);
    }

    if (Num_Snd_Grid_CA2CB[IDS]!=0)  MPI_Wait(&request,&stat);
  }

  /* reset: Index_Snd_Grid_CA2CB

     Index_Snd_Grid_CB2CA:  BN_CA
     Index_Rcv_Grid_CB2CA:  BN_CB
  */

  for (ID=0; ID<numprocs; ID++) Num_Snd_Grid_CA2CB[ID] = 0;

  N2D = OFFT_Ngrid3*OFFT_Ngrid2; /* for CB */

  for (BN_CA=0; BN_CA<My_NumGrid_CA; BN_CA++){

    GN_CA = GNs + BN_CA;    
    n3 = GN_CA/(OFFT_Ngrid1*OFFT_Ngrid2);
    n1 = (GN_CA - n3*(OFFT_Ngrid1*OFFT_Ngrid2))/OFFT_Ngrid2;
    n2 = GN_CA - n3*(OFFT_Ngrid1*OFFT_Ngrid2) - n1*OFFT_Ngrid2;
    n2D = n3*OFFT_Ngrid2 + n2;
    ID = n2D*numprocs/N2D;
    Index_Snd_Grid_CA2CB[ID][Num_Snd_Grid_CA2CB[ID]] = BN_CA;
    Num_Snd_Grid_CA2CB[ID]++;
  }

  /* find the maximum Num_Snd_Grid_CA2CB and Num_Rcv_Grid_CA2CB */

  Max_Num_Snd_Grid_CA2CB = 0;
  Max_Num_Rcv_Grid_CA2CB = 0;

  for (ID=0; ID<numprocs; ID++){
    if (Max_Num_Snd_Grid_CA2CB<Num_Snd_Grid_CA2CB[ID]){ 
      Max_Num_Snd_Grid_CA2CB = Num_Snd_Grid_CA2CB[ID];
    }

    if (Max_Num_Rcv_Grid_CA2CB<Num_Rcv_Grid_CA2CB[ID]){ 
      Max_Num_Rcv_Grid_CA2CB = Num_Rcv_Grid_CA2CB[ID];
    }
  }

  /* find NN_CA2CB_S and NN_CA2CB_R 
     and set ID_NN_CA2CB_S, 
             ID_NN_CA2CB_R,
             GP_CA2CB_S,
             GP_CA2CB_R
  */

  NN_CA2CB_S = 0;
  NN_CA2CB_R = 0;

  for (ID=0; ID<numprocs; ID++){
    IDS = (myid + ID) % numprocs;
    IDR = (myid - ID + numprocs) % numprocs;
    if (Num_Snd_Grid_CA2CB[IDS]!=0) NN_CA2CB_S++;
    if (Num_Rcv_Grid_CA2CB[IDR]!=0) NN_CA2CB_R++;
  }

  ID_NN_CA2CB_S = (int*)malloc(sizeof(int)*NN_CA2CB_S);
  ID_NN_CA2CB_R = (int*)malloc(sizeof(int)*NN_CA2CB_R);
  GP_CA2CB_S = (int*)malloc(sizeof(int)*(NN_CA2CB_S+1));
  GP_CA2CB_R = (int*)malloc(sizeof(int)*(NN_CA2CB_R+1));

  NN_CA2CB_S = 0;
  NN_CA2CB_R = 0;
  GP_CA2CB_S[0] = 0;
  GP_CA2CB_R[0] = 0;

  for (ID=0; ID<numprocs; ID++){

    IDS = (myid + ID) % numprocs;
    IDR = (myid - ID + numprocs) % numprocs;

    if (Num_Snd_Grid_CA2CB[IDS]!=0){
      ID_NN_CA2CB_S[NN_CA2CB_S] = IDS;
      NN_CA2CB_S++;
      GP_CA2CB_S[NN_CA2CB_S] = GP_CA2CB_S[NN_CA2CB_S-1] + Num_Snd_Grid_CA2CB[IDS];
    }

    if (Num_Rcv_Grid_CA2CB[IDR]!=0){
      ID_NN_CA2CB_R[NN_CA2CB_R] = IDR;
      NN_CA2CB_R++;
      GP_CA2CB_R[NN_CA2CB_R] = GP_CA2CB_R[NN_CA2CB_R-1] + Num_Rcv_Grid_CA2CB[IDR];
    }
  }

  /* set My_NumGrid_CB */

  N2D = OFFT_Ngrid3*OFFT_Ngrid2;
  My_NumGrid_CB = (((myid+1)*N2D+numprocs-1)/numprocs)*OFFT_Ngrid1
                  - ((myid*N2D+numprocs-1)/numprocs)*OFFT_Ngrid1;
  OFFT_My_NumGrid_Out = My_NumGrid_CB;

  /* set myStart,myEnd,OFFT_My_Index_Out */

  N2D = OFFT_Ngrid3*OFFT_Ngrid2;
  myStart = ((myid*N2D+numprocs-1)/numprocs)*OFFT_Ngrid1;
  myEnd   = (((myid+1)*N2D+numprocs-1)/numprocs)*OFFT_Ngrid1 - 1;

  N2D = OFFT_Ngrid2*OFFT_Ngrid1;
  My_Index_Out[0] = myStart/N2D;
  My_Index_Out[1] = (myStart-My_Index_Out[0]*N2D)/OFFT_Ngrid1;
  My_Index_Out[2] = myStart-My_Index_Out[0]*N2D-My_Index_Out[1]*OFFT_Ngrid1;
  
  My_Index_Out[3] = myEnd/N2D;
  My_Index_Out[4] = (myEnd-My_Index_Out[3]*N2D)/OFFT_Ngrid1;
  My_Index_Out[5] = myEnd-My_Index_Out[3]*N2D-My_Index_Out[4]*OFFT_Ngrid1;

  /* find My_Max_NumGrid */

  OFFT_My_Max_NumGrid = 0;
  if (OFFT_My_Max_NumGrid<My_NumGrid_AB) OFFT_My_Max_NumGrid = My_NumGrid_AB;
  if (OFFT_My_Max_NumGrid<My_NumGrid_CA) OFFT_My_Max_NumGrid = My_NumGrid_CA;
  if (OFFT_My_Max_NumGrid<My_NumGrid_CB) OFFT_My_Max_NumGrid = My_NumGrid_CB;

  *My_Max_NumGrid = OFFT_My_Max_NumGrid;
  *My_NumGrid_In = OFFT_My_NumGrid_In;
  *My_NumGrid_Out = OFFT_My_NumGrid_Out;

  if(OFFT_measure_time){
      MPI_Barrier(mpi_comm_level1);
      openfft_dtime(&TEtime);
      time0 = TEtime - TStime;
  }

  /* Print memory usage */

  if(OFFT_print_memory==1 && myid==0){
    MB = 1024*1024;
    size_array0 = GP_AB2CA_S[NN_AB2CA_S] > GP_CA2CB_S[NN_CA2CB_S] ? 
      GP_AB2CA_S[NN_AB2CA_S] : GP_CA2CB_S[NN_CA2CB_S];
    size_array1 = GP_AB2CA_R[NN_AB2CA_R] > GP_CA2CB_R[NN_CA2CB_R] ? 
      GP_AB2CA_R[NN_AB2CA_R] : GP_CA2CB_R[NN_CA2CB_R];

    printf("=========OpenFFT: Memory Usage per Process============\n");
    printf("Index_Snd_Grid_AB2CA: %10.2f MB\n",(double)(sizeof(int)*size_Index_Snd_Grid_AB2CA/MB));
    printf("Index_Rcv_Grid_AB2CA: %10.2f MB\n",(double)(sizeof(int)*size_Index_Rcv_Grid_AB2CA/MB));
    printf("Index_Snd_Grid_CA2CB: %10.2f MB\n",(double)(sizeof(int)*size_Index_Snd_Grid_CA2CB/MB));
    printf("Index_Rcv_Grid_CA2CB: %10.2f MB\n",(double)(sizeof(int)*size_Index_Rcv_Grid_CA2CB/MB));
    printf("ID_NN_AB2CA_S       : %10.2f MB\n",(double)(sizeof(int)*NN_AB2CA_S/MB));
    printf("ID_NN_AB2CA_R       : %10.2f MB\n",(double)(sizeof(int)*NN_AB2CA_R/MB));
    printf("ID_NN_CA2CB_S       : %10.2f MB\n",(double)(sizeof(int)*NN_CA2CB_S/MB));
    printf("ID_NN_CA2CB_R       : %10.2f MB\n",(double)(sizeof(int)*NN_CA2CB_R/MB));
    printf("GP_AB2CA_S          : %10.2f MB\n",(double)(sizeof(int)*(NN_AB2CA_S+1)/MB));
    printf("GP_AB2CA_R          : %10.2f MB\n",(double)(sizeof(int)*(NN_AB2CA_R+1)/MB));
    printf("GP_CA2CB_S          : %10.2f MB\n",(double)(sizeof(int)*(NN_CA2CB_S+1)/MB));
    printf("GP_CA2CB_R          : %10.2f MB\n",(double)(sizeof(int)*(NN_CA2CB_R+1)/MB));
    printf("Temporary array0    : %10.2f MB\n",(double)(sizeof(dcomplex)*size_array0/MB));
    printf("Temporary array1    : %10.2f MB\n",(double)(sizeof(dcomplex)*size_array1/MB));
    printf("Temporary OFFT_in   : %10.2f MB\n",(double)(sizeof(fftw_complex)*Max_OneD_Grids/MB));
    printf("Temporary OFFT_out  : %10.2f MB\n",(double)(sizeof(fftw_complex)*Max_OneD_Grids/MB));
    printf("=====================================================\n");
  }

  return time0;
}  


